@@ -14,12 +14,14 @@ Notes:
 from controlSBML.make_roadrunner import makeRoadrunner
 import controlSBML as ctl
 from controlSBML import util
+from controlSBML import msgs
 
 import control
 import numpy as np
 import pandas as pd
 
 
+ATOL = 1e-8  # Absolute tolerance for comparisons
 TIME = "time"
 START_TIME = 0  # Default start time
 END_TIME = 5  # Default endtime
@@ -47,12 +49,23 @@ class ControlBase(object):
         is_reduced: bool
             construct a reduced model so that the A matrix is nonsingular
         """
-        self.is_reduced = is_reduced
-        # Iinitial model calculations
+        # First initializations
         self.model_reference = model_reference
         self.roadrunner = makeRoadrunner(self.model_reference)
         self.species_names = list(
               self.roadrunner.getFullStoichiometryMatrix().rownames)
+        if output_names is None:
+            output_names = self.species_names
+        self.output_names = output_names
+        # Initializations with error checks
+        self.is_reduced = is_reduced
+        if len(set(self.roadrunner.getReactionIds()).intersection(self.output_names)) > 0:
+            if self.is_reduced:
+                self.is_reduced = False
+                text = "Cannot have a flux output and is_reduced=True."
+                text += "  Setting is_reduced=False."
+                msgs.warn(text)
+        # Iinitial model calculations
         self._jacobian_time = TIMEPOINT_NULL
         self._jacobian_df = None
         # Set defaults.
@@ -63,7 +76,10 @@ class ControlBase(object):
         self.full_stoichiometry_df, self.reduced_stoichiometry_df  \
               = self._makeStoichiometryDF()
         # Check for consistency on the state specification
-        self.reaction_names = list(self.reduced_stoichiometry_df.columns)
+        if self.is_reduced:
+            self.reaction_names = list(self.reduced_stoichiometry_df.columns)
+        else:
+            self.reaction_names = list(self.reduced_stoichiometry_df.columns)
         # Handle defaults
         if input_names is None:
             self.input_names = []
@@ -71,10 +87,8 @@ class ControlBase(object):
             self.input_names = input_names
         #self.input_names = self._sortList(self.reaction_names, self.input_names)
         self.num_input = len(self.input_names)
-        if output_names is None:
-            output_names = self.species_names
-        self.output_names = output_names
-        self.output_names = self._sortList(self.species_names, self.output_names)
+        # FIXME: Is this necessary?
+        # self.output_names = self._sortList(self.species_names, self.output_names)
         self.num_output = len(self.output_names)
         # Other calculations
         self.antimony = self.roadrunner.getAntimony()
@@ -86,14 +100,15 @@ class ControlBase(object):
             text += "  Species are: %s" % str(self.species_names)
             text += "  States are: %s" % str(self.state_names)
             raise RuntimeError(text)
-        if not set(self.output_names) <= set(self.species_names):
+        possible_names = set(self.species_names).union(self.reaction_names)
+        if not set(self.output_names) <= set(possible_names):
             diff = list(set(self.output_names).difference(self.species_names))
-            text = "Outputs must be species. The following outputs are not species"
-            text += "The following outputs are not species: %s" % str(diff)
+            text = "Outputs must be species or fluxes."
+            text += "The following outputs are invalid: %s" % str(diff)
             raise ValueError(text)
-        possible_input_names = set(self.state_names).union(self.reaction_names)
-        if not set(self.input_names) <= set(possible_input_names):
-            diff = list(set(self.input_names).difference(possible_input_names))
+        possible_names = set(self.species_names).union(self.reaction_names)
+        if not set(self.input_names) <= set(possible_names):
+            diff = list(set(self.input_names).difference(possible_names))
             text = "Inputs must be a species or a reaction."
             text += "   Invalid names are: %s" % str(diff)
             raise ValueError(text)
@@ -126,23 +141,51 @@ class ControlBase(object):
         """
         Creates the output C dataframe based on the requested output_names.
         Columns should be the states. Rows are the outputs.
+        There are 3 cases for outputs:
+         1) The output is a state
+         2) The output is a floating species whose concentration
+            is a linear function of the other state variables
+         3) The output is a flux and this
+            is a linear function of the other state variables
 
         Returns
         -------
         pd.DataFrame
         """
+        # FIXME: Not correctly constructing the C matrix
+        # Initializations 
         state_names = self.state_names
-        C_df = pd.DataFrame(np.eye(self.num_state), columns=state_names,
-              index=state_names)
-        if self.is_reduced:
-            L0 = self.roadrunner.getL0Matrix()
-            if len(L0) > 0:
-                L0_df = pd.DataFrame(L0, columns=L0.colnames, index=L0.rownames)
-                C_df = pd.concat([C_df, L0_df], axis=0)
-        C_df = C_df.loc[self.output_names, :]
+        num_output = len(self.output_names)
+        if len(set(self.reaction_names).intersection(self.output_names)) > 0:
+            flux_jacobian_df = self.makeFluxJacobian()
+        else:
+            flux_jacobian_df = None
+        L0 = self.roadrunner.getL0Matrix()
+        if len(L0) > 0:
+            L0_df = pd.DataFrame(L0, columns=L0.colnames, index=L0.rownames)
+        else:
+            L0_df = None
+        # Iterate across each output to construct the transpose of the C matrix
+        C_T_dct = {}
+        for name in self.output_names:
+            if name in self.state_names:
+                values = np.repeat(0, self.num_state)
+                idx = self.state_names.index(name)
+                values[idx] = 1
+                C_T_dct[name] = values
+            elif name in self.species_names:
+                if L0_df is None:
+                    raise RuntimeError("Species missing from L0")
+                if not name in L0.rownames:
+                    raise RuntimeError("Species missing from L0")
+                C_T_dct[name] = L0_df.loc[name, :]
+            elif name in self.reaction_names:
+                C_T_dct[name] = flux_jacobian_df.loc[name, :]
+        C_df_T = pd.DataFrame(C_T_dct, index=state_names)
+        C_df = C_df_T.transpose()
         values = C_df.values.flatten()
         if any([np.isnan(v) for v in values]):
-            import pdb; pdb.set_trace()
+            raise RuntimeError("Nan value encountered.")
         return C_df
 
     @property
@@ -477,7 +520,7 @@ class ControlBase(object):
         return ctl.NonlinearIOSystem(name, self, effector_dct=effector_dct)
 
     @staticmethod
-    def reduceTransferFunction(tf):
+    def reduceTransferFunction(tf, atol=ATOL):
         """
         Reduces the order of a transfer function if trailing zeroes.
 
@@ -492,7 +535,7 @@ class ControlBase(object):
         def findOrderOfFirstNonzeroDigit(polynomial):
             for idx in range(len(polynomial)):
                 pos = len(polynomial) - idx - 1
-                if not np.isclose(polynomial[pos], 0):
+                if not np.isclose(polynomial[pos], 0, atol=atol):
                     return idx
             return len(polynomial) - 1
         def reduceOrder(polynomial, new_order):
@@ -509,7 +552,7 @@ class ControlBase(object):
         new_tf = control.TransferFunction(new_numerator, new_denominator)
         return new_tf
 
-    def makeTransferFunction(self, time=None):
+    def makeTransferFunction(self, time=None, atol=ATOL):
         """
         Creates a transfer function for the system. Verifies that there
         is a single input and a single output. Reduces the order of the
@@ -518,6 +561,7 @@ class ControlBase(object):
         Parameters
         ----------
         time: float (time at which Jacobian is obtained)
+        atol: absolute tolerance for comparison
         
         Returns
         -------
@@ -532,4 +576,44 @@ class ControlBase(object):
         state_space = self.makeStateSpace(time=time)
         tf = control.ss2tf(state_space)
         #
-        return self.reduceTransferFunction(tf)
+        return self.reduceTransferFunction(tf, atol=atol)
+
+    def makeFluxJacobian(self, time=None):
+        """
+        Constructs the Jacobian of the reaction flux vector.
+
+        Parameters
+        ----------
+        time: float
+            Time at which this is calculated
+        
+        Returns
+        -------
+        pd.DataFrame
+            index: reaction id
+            column: species
+        """
+        # FIXME : Handle case where state_names < species_names
+        # This calculation only works if state_names == species_names
+        diff = set(self.state_names).symmetric_difference(self.species_names)
+        if len(diff) > 0:
+            raise RuntimeError(
+                  "Code doesn't work if species_names != state_names")
+        # Adjust time if necessary
+        cur_time = self.getTime()
+        if time is None:
+            time = cur_time
+        if not np.isclose(time, cur_time):
+            self.setTime(time)
+        #
+        dct = {}
+        for state_name in self.state_names:
+            dct[state_name] = []
+            for reaction_name in self.reaction_names:
+                dct[state_name].append(self.roadrunner.getEE(
+                      reaction_name, state_name))
+        #
+        df = pd.DataFrame(dct, index=self.reaction_names)
+        if time != cur_time:
+            self.setTime(cur_time)
+        return df
